""" Abstract Controller Definition """

from abc import ABC, abstractmethod
from typing import Any, Optional

from ..services.auth import AuthService


class AbstractController(ABC):
    """
    Abstract Controller

    Attributes
    ----------
    auth_service: AuthService
        The Auth Service.
    """

    auth_service: AuthService

    def __init__(self, auth_service: AuthService):
        self.auth_service = auth_service

    @abstractmethod
    def execute(self, *args, **kwargs) -> Optional[Any]:
        """Should be implemented in the subclass"""



""" Auth Service User Base Definition """

from typing import Union

from pydantic import BaseModel


class UserBase(BaseModel):
    """
    Auth Service User Base

    Attributes
    ----------
    username: str
        The username of the user.
    guid: str
        The user guid.
    email: Union[str, None] = None
        The user email address.
    disabled: Union[bool, None] = None
        Whether the account is enabled or not.
    admin: Union[bool, None] = None
        Whether the account is an admin or not.
    """

    username: str
    guid: str
    email: Union[str, None] = None
    disabled: Union[bool, None] = None
    admin: Union[bool, None] = None



""" DB DAO Utilities """

import logging
import socket

from mysql.connector import Error, errorcode
from mysql.connector.pooling import MySQLConnectionPool

from rowantree.common.sdk import demand_env_var


# pylint: disable=duplicate-code
def get_connect_pool() -> MySQLConnectionPool:
    """
    Gets a connection pool to the database.

    Returns
    -------
    cnxpool: MySQLConnectionPool
        A connection pool to the database server.
    """

    try:
        logging.debug("Connecting to database")
        cnxpool: MySQLConnectionPool = MySQLConnectionPool(
            pool_name="servercnxpool",
            pool_size=3,
            user=demand_env_var(name="DATABASE_USERNAME"),
            password=demand_env_var(name="DATABASE_PASSWORD"),
            host=demand_env_var(name="DATABASE_SERVER"),
            database=demand_env_var(name="DATABASE_NAME"),
        )
        return cnxpool
    except socket.error as error:
        logging.debug(error)
        raise error
    except Error as error:
        if error.errno == errorcode.ER_ACCESS_DENIED_ERROR:
            logging.debug("Something is wrong with your user name or password")
        elif error.errno == errorcode.ER_BAD_DB_ERROR:
            logging.debug("Database does not exist")
        else:
            logging.debug(error)
        raise error



""" Abstract Controller Definition """

from pydantic import BaseModel

from .db.dao import DBDAO


class AbstractService(BaseModel):
    """
    Abstract Service

    Attributes
    ----------
    dao: DBDAO
        The database DAO.
    """

    dao: DBDAO

    class Config:
        """
        Pydantic Default Configuration Over-Ride
        """

        arbitrary_types_allowed = True



""" Token Controller Definition """

from typing import Optional

from fastapi.security import OAuth2PasswordRequestForm
from starlette import status
from starlette.exceptions import HTTPException

from rowantree.auth.sdk.contracts.dto.token import Token

from ..contracts.dto.user.user import User
from ..services.auth import AuthService
from .abstract_controller import AbstractController


class TokenController(AbstractController):
    """Token Controller"""

    def execute(self, request: OAuth2PasswordRequestForm) -> Token:
        """
        Execute the controller action. (Auth User)

        Parameters
        ----------
        request: OAuth2PasswordRequestForm
            The username/password form data

        Returns
        -------
        token: Token
            A token with the users permissions.
        """

        user: Optional[User] = self.auth_service.authenticate_user(request.username, request.password)

        if not user:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return AuthService.create_user_access_token(user=user)



""" Database DAO Definition """

import logging
import socket
from typing import Optional, Tuple

import mysql.connector
from mysql.connector import errorcode
from mysql.connector.pooling import MySQLConnectionPool
from starlette import status
from starlette.exceptions import HTTPException

from ...contracts.dto.user.user import User
from .incorrect_row_count_error import IncorrectRowCountError


class DBDAO:
    """
    Database DAO

    Attributes
    ----------
    cnxpool: Any
        MySQL Connection Pool
    """

    cnxpool: MySQLConnectionPool

    def __init__(self, cnxpool: MySQLConnectionPool):
        self.cnxpool = cnxpool

    # pylint: disable=duplicate-code
    def _call_proc(self, name: str, args: list, debug: bool = False) -> Optional[list[Tuple]]:
        """
        Perform a stored procedure call.

        Parameters
        ----------
        name: str
            The name of the stored procedure to call.
        args: list
            The arguments to pass to the stored procedure.
        debug: bool
            Whether to log debug details about the call.

        Returns
        -------
        results: Optional[list[Tuple]]
            An optional list of tuples (rows) from the call.
        """

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Name: {%s}, Arguments: {%s}", name, args)
        rows: Optional[list[Tuple]] = None
        try:
            cnx = self.cnxpool.get_connection()
            cursor = cnx.cursor()
            cursor.callproc(name, args)
            for result in cursor.stored_results():
                rows = result.fetchall()
            cursor.close()
        except socket.error as error:
            logging.debug(error)
            raise error
        except mysql.connector.Error as error:
            if error.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                logging.debug("Something is wrong with your user name or password")
            elif error.errno == errorcode.ER_BAD_DB_ERROR:
                logging.debug("Database does not exist")
            else:
                logging.debug(error)
            raise error
        else:
            cnx.close()

        if debug:
            logging.debug("[DAO] [Stored Proc Call Details] Returning:")
            logging.debug(rows)
        return rows

    def get_user(self, username: Optional[str] = None, guid: Optional[str] = None) -> User:
        """
        Get User (From Database) Using `username` first, else `guid`.  At least one must be supplied.

        Parameters
        ----------
        username: Optional[str]
            username
        guid: Optional[str]
            user guid

        Returns
        -------
        user: User
            User (from database)
        """

        if username is None and guid is None:
            logging.debug("get_user called without username or a guid, one is required")
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        args: list[str] = [username]
        proc_name: str = "getUserByUsername"

        if username is None:
            args = [guid]
            proc_name: str = "getUserByGUID"

        rows: list[Tuple] = self._call_proc(proc_name, args, True)
        if len(rows) != 1:
            raise IncorrectRowCountError(f"Result count was not exactly one. Received: {rows}")
        user: tuple = rows[0]

        is_disabled: bool = True
        if user[6] == 0:
            is_disabled = False

        is_admin: bool = False
        if user[7] == 1:
            is_admin = True

        return User(
            username=user[2], guid=user[1], email=user[3], hashed_password=user[4], disabled=is_disabled, admin=is_admin
        )



""" The Rowan Tree Auth Service Definition """

import logging
import os
from pathlib import Path

from fastapi import Depends, FastAPI, status
from fastapi.security import OAuth2PasswordRequestForm
from mysql.connector.pooling import MySQLConnectionPool
from starlette.middleware.cors import CORSMiddleware

from rowantree.auth.sdk.contracts.dto.token import Token
from rowantree.common.sdk import demand_env_var

from ..controllers.token import TokenController
from ..services.auth import AuthService
from ..services.db.dao import DBDAO
from ..services.db.utils import get_connect_pool

# Setup logging
Path(demand_env_var(name="LOGS_DIR")).mkdir(parents=True, exist_ok=True)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%m/%d/%Y %I:%M:%S %p",
    level=logging.DEBUG,
    filemode="w",
    filename=f"{demand_env_var(name='LOGS_DIR')}/{os.uname()[1]}.therowantree.auth.service.log",
)

logging.debug("Starting server")

# Creating database connection pool, and DAO
cnxpool: MySQLConnectionPool = get_connect_pool()
dao: DBDAO = DBDAO(cnxpool=cnxpool)
auth_service: AuthService = AuthService(dao=dao)


# Create controllers
token_controller: TokenController = TokenController(auth_service=auth_service)

# Create the FastAPI application
app = FastAPI()

#  Apply COR Configuration | https://fastapi.tiangolo.com/tutorial/cors/
origins = ["*"]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


# Define our handlers


@app.get("/health/plain", status_code=status.HTTP_200_OK)
def health_plain() -> bool:
    """
    Get Application Health
    [GET] /health/plain

    Returns
    -------
    health: bool
        A true/false response of server health.

    [STATUS CODE] 200: OK
    [STATUS CODE] 400: Invalid Request
    [STATUS CODE] 500: Internal Server Error
    """

    return True


@app.post("/v1/auth/token", status_code=status.HTTP_200_OK)
def token_handler(form_data: OAuth2PasswordRequestForm = Depends()) -> Token:
    """
    Get Token
    [POST] /v1/auth/token

    Returns
    -------
    token: Token

    [STATUS CODE] 200: OK
    [STATUS CODE] 400: Invalid Request
    [STATUS CODE] 401: Permission Denied
    [STATUS CODE] 500: Internal Server Error
    """

    return token_controller.execute(request=form_data)



""" Authentication and Authorization Service Definition """

import logging
from datetime import datetime, timedelta
from typing import Optional

from jose import jwt
from passlib.context import CryptContext

from rowantree.auth.sdk.contracts.dto.token import Token
from rowantree.common.sdk import demand_env_var, demand_env_var_as_float

from ..contracts.dto.user.user import User
from .abstract_service import AbstractService
from .db.incorrect_row_count_error import IncorrectRowCountError


class AuthService(AbstractService):
    """
    Authentication and Authorization Service

    Attributes
    ----------
    pwd_context: CryptContext = CryptContext(schemes=["bcrypt"], deprecated="auto")
        Password hashing context
    """

    pwd_context: CryptContext = CryptContext(schemes=["bcrypt"], deprecated="auto")

    def verify_password(self, plain_password: str, hashed_password: str) -> bool:
        """
        Verifies hashed password.

        Parameters
        ----------
        plain_password: str
            Plan text password
        hashed_password: str
            Hashed password

        Returns
        -------
        verified: bool
            True if the password matches the hash, False otherwise.
        """

        return self.pwd_context.verify(plain_password, hashed_password)

    def get_password_hash(self, password: str) -> str:
        """
        Generates password hash.

        Parameters
        ----------
        password: str
            Plain text password.

        Returns
        -------
        hashed_password: str
            Hashed password.
        """

        return self.pwd_context.hash(password)

    def authenticate_user(self, username: str, password: str) -> Optional[User]:
        """
        Authenticates a user by username/password.

        Parameters
        ----------
        username: str
            Username
        password:
            Plain text password.

        Returns
        -------
        user: Optional[User]
            Returns an instance of `UserInDB` if successfully authenticated, otherwise `None`.
        """

        user: Optional[User] = None

        try:
            user = self.dao.get_user(username=username)
        except IncorrectRowCountError as error:
            logging.debug("User not found: %s", str(error))
        if not user:
            return None
        if not self.verify_password(password, user.hashed_password):
            return None

        return user

    @staticmethod
    def create_user_access_token(user: User) -> Token:
        """
        Create bearer Token from user instance.

        Parameters
        ----------
        user: User
            An instance of a UserInDB object.

        Returns
        -------
        token: Token
            A bearer token for the requested user.
        """

        # Currently these are the claims we pull from the database.
        data: dict = {"sub": user.guid, "disabled": user.disabled, "admin": user.admin}
        return AuthService.create_access_token(data=data)

    @staticmethod
    def create_access_token(data: dict) -> Token:
        """
        Mints the access token.

        Parameters
        ----------
        data: dict
            The base set of claims to include in the token.

        Returns
        -------
        token: Token
            A newly minted token.
        """

        to_encode: dict = data.copy()
        expire: datetime = datetime.utcnow() + timedelta(
            minutes=demand_env_var_as_float(name="ACCESS_TOKEN_EXPIRATION_TIME")
        )
        to_encode.update({"iss": demand_env_var(name="ACCESS_TOKEN_ISSUER"), "exp": expire})
        encoded_jwt: str = jwt.encode(
            to_encode,
            demand_env_var(name="ACCESS_TOKEN_SECRET_KEY"),
            algorithm=demand_env_var(name="ACCESS_TOKEN_ALGORITHM"),
        )
        return Token(access_token=encoded_jwt, token_type="bearer")

    # TODO: Future functionality.
    # def get_user_by_jwt(self, token: str) -> User:
    #     credentials_exception: HTTPException = HTTPException(
    #         status_code=status.HTTP_401_UNAUTHORIZED,
    #         detail="Could not validate credentials",
    #         headers={"WWW-Authenticate": "Bearer"},
    #     )
    #     user: Optional[User] = None
    #     token_claims: TokenClaims = get_claims(token=token)
    #
    #     try:
    #         user = self.dao.get_user(guid=token_claims.sub)
    #     except IncorrectRowCountError as error:
    #         logging.debug(f"User not found: {str(error)}")
    #
    #     if user is None:
    #         raise credentials_exception
    #
    #     return user
